#define _WIN32_WINNT 0x501
#include <windows.h>
#include <winternl.h>
#include <specstrings.h>
#include <cstdio>

// http://support.microsoft.com/kb/131065/en-us
BOOL SetPrivilege(
				  HANDLE hToken,          // token handle
				  LPCTSTR Privilege,      // Privilege to enable/disable
				  BOOL bEnablePrivilege   // TRUE to enable.  FALSE to disable
				  );

namespace SysDbg
{
	enum SYSDBG_COMMAND
	{
		SysDbgGetVersion = 7,
		SysDbgReadMemory = 8,
	};

	struct MEMORY_QUERY
	{
		DWORD Address;
		PVOID Data;
		ULONG Length;
	};

	typedef NTSYSAPI NTSTATUS ( NTAPI * NtSystemDebugControlFunc ) (
		__in SYSDBG_COMMAND Command,
		__in PVOID InputBuffer,
		__in ULONG InputBufferLength,
		__out PVOID OutputBuffer,
		__in ULONG OutputBufferLength,
		__out PULONG ReturnLength
		);

	// dt nt!_DBGKD_GET_VERSION64 @windbg
	struct DBGKD_GET_VERSION64
	{
		USHORT  MajorVersion;
		USHORT  MinorVersion;
		USHORT  ProtocolVersion;
		USHORT  Flags;
		USHORT  MachineType;
		UCHAR   MaxPacketType;
		UCHAR   MaxStateChange;
		UCHAR   MaxManipulate;
		UCHAR   Simulation;
		USHORT  Unused[1];
		ULONG64 KernBase;
		ULONG64 PsLoadedModuleList;
		ULONG64 DebuggerDataList;
	};

	// copied from "WDbgExts.h"
	typedef struct _DBGKD_DEBUG_DATA_HEADER64
	{
		LIST_ENTRY64 List;
		ULONG        OwnerTag;
		ULONG        Size;
	} DBGKD_DEBUG_DATA_HEADER64, *PDBGKD_DEBUG_DATA_HEADER64;

	// copied from "WDbgExts.h"
	// this struct was modified;
	// the original version has more members.
	typedef struct _KDDEBUGGER_DATA64
	{

		DBGKD_DEBUG_DATA_HEADER64 Header;

		//
		// Base address of kernel image
		//

		ULONG64   KernBase;

		//
		// DbgBreakPointWithStatus is a function which takes an argument
		// and hits a breakpoint.  This field contains the address of the
		// breakpoint instruction.  When the debugger sees a breakpoint
		// at this address, it may retrieve the argument from the first
		// argument register, or on x86 the eax register.
		//

		ULONG64   BreakpointWithStatus;       // address of breakpoint

		//
		// Address of the saved context record during a bugcheck
		//
		// N.B. This is an automatic in KeBugcheckEx's frame, and
		// is only valid after a bugcheck.
		//

		ULONG64   SavedContext;

		//
		// help for walking stacks with user callbacks:
		//

		//
		// The address of the thread structure is provided in the
		// WAIT_STATE_CHANGE packet.  This is the offset from the base of
		// the thread structure to the pointer to the kernel stack frame
		// for the currently active usermode callback.
		//

		USHORT  ThCallbackStack;            // offset in thread data

		//
		// these values are offsets into that frame:
		//

		USHORT  NextCallback;               // saved pointer to next callback frame
		USHORT  FramePointer;               // saved frame pointer

		//
		// pad to a quad boundary
		//
		USHORT  PaeEnabled:1;

		//
		// Address of the kernel callout routine.
		//

		ULONG64   KiCallUserMode;             // kernel routine

		//
		// Address of the usermode entry point for callbacks.
		//

		ULONG64   KeUserCallbackDispatcher;   // address in ntdll


		//
		// Addresses of various kernel data structures and lists
		// that are of interest to the kernel debugger.
		//

		ULONG64   PsLoadedModuleList;
		ULONG64   PsActiveProcessHead;
		ULONG64   PspCidTable;

		ULONG64   ExpSystemResourcesList;
		ULONG64   ExpPagedPoolDescriptor;
		ULONG64   ExpNumberOfPagedPools;

		ULONG64   KeTimeIncrement;
		ULONG64   KeBugCheckCallbackListHead;
		ULONG64   KiBugcheckData;

		ULONG64   IopErrorLogListHead;

		ULONG64   ObpRootDirectoryObject;
		ULONG64   ObpTypeObjectType;

		ULONG64   MmSystemCacheStart;
		ULONG64   MmSystemCacheEnd;
		ULONG64   MmSystemCacheWs;

		ULONG64   MmPfnDatabase;
		ULONG64   MmSystemPtesStart;
		ULONG64   MmSystemPtesEnd;
		ULONG64   MmSubsectionBase;
		ULONG64   MmNumberOfPagingFiles;

		ULONG64   MmLowestPhysicalPage;
		ULONG64   MmHighestPhysicalPage;
		ULONG64   MmNumberOfPhysicalPages;

		ULONG64   MmMaximumNonPagedPoolInBytes;
		ULONG64   MmNonPagedSystemStart;
		ULONG64   MmNonPagedPoolStart;
		ULONG64   MmNonPagedPoolEnd;

		ULONG64   MmPagedPoolStart;
		ULONG64   MmPagedPoolEnd;
		ULONG64   MmPagedPoolInformation;
		ULONG64   MmPageSize;

		ULONG64   MmSizeOfPagedPoolInBytes;

		ULONG64   MmTotalCommitLimit;
		ULONG64   MmTotalCommittedPages;
		ULONG64   MmSharedCommit;
		ULONG64   MmDriverCommit;
		ULONG64   MmProcessCommit;
		ULONG64   MmPagedPoolCommit;
		ULONG64   MmExtendedCommit;

		ULONG64   MmZeroedPageListHead;
		ULONG64   MmFreePageListHead;
		ULONG64   MmStandbyPageListHead;
		ULONG64   MmModifiedPageListHead;
		ULONG64   MmModifiedNoWritePageListHead;
		ULONG64   MmAvailablePages;
		ULONG64   MmResidentAvailablePages;

		ULONG64   PoolTrackTable;
		ULONG64   NonPagedPoolDescriptor;

		ULONG64   MmHighestUserAddress;
		ULONG64   MmSystemRangeStart;
		ULONG64   MmUserProbeAddress;

		ULONG64   KdPrintCircularBuffer;
		ULONG64   KdPrintCircularBufferEnd;
		ULONG64   KdPrintWritePointer;
		ULONG64   KdPrintRolloverCount;

		ULONG64   MmLoadedUserImageList;
		// NT 5.1 Addition

		ULONG64   NtBuildLab;
		ULONG64   KiNormalSystemCall;

		// NT 5.0 QFE addition

		ULONG64   KiProcessorBlock;
		ULONG64   MmUnloadedDrivers;
		ULONG64   MmLastUnloadedDriver;
		ULONG64   MmTriageActionTaken;
		ULONG64   MmSpecialPoolTag;
		ULONG64   KernelVerifier;
		ULONG64   MmVerifierData;
		ULONG64   MmAllocatedNonPagedPool;
		ULONG64   MmPeakCommitment;
		ULONG64   MmTotalCommitLimitMaximum;
		ULONG64   CmNtCSDVersion;

		// NT 5.1 Addition

		ULONG64   MmPhysicalMemoryBlock;
		ULONG64   MmSessionBase;
		ULONG64   MmSessionSize;
		ULONG64   MmSystemParentTablePage;

#if 0
		// Server 2003 addition

		ULONG64   MmVirtualTranslationBase;

		USHORT    OffsetKThreadNextProcessor;
		USHORT    OffsetKThreadTeb;
		USHORT    OffsetKThreadKernelStack;
		USHORT    OffsetKThreadInitialStack;

		USHORT    OffsetKThreadApcProcess;
		USHORT    OffsetKThreadState;
		USHORT    OffsetKThreadBStore;
		USHORT    OffsetKThreadBStoreLimit;

		USHORT    SizeEProcess;
		USHORT    OffsetEprocessPeb;
		USHORT    OffsetEprocessParentCID;
		USHORT    OffsetEprocessDirectoryTableBase;

		USHORT    SizePrcb;
		USHORT    OffsetPrcbDpcRoutine;
		USHORT    OffsetPrcbCurrentThread;
		USHORT    OffsetPrcbMhz;

		USHORT    OffsetPrcbCpuType;
		USHORT    OffsetPrcbVendorString;
		USHORT    OffsetPrcbProcStateContext;
		USHORT    OffsetPrcbNumber;

		USHORT    SizeEThread;

		ULONG64   KdPrintCircularBufferPtr;
		ULONG64   KdPrintBufferSize;

		ULONG64   KeLoaderBlock;

		USHORT    SizePcr;
		USHORT    OffsetPcrSelfPcr;
		USHORT    OffsetPcrCurrentPrcb;
		USHORT    OffsetPcrContainedPrcb;

		USHORT    OffsetPcrInitialBStore;
		USHORT    OffsetPcrBStoreLimit;
		USHORT    OffsetPcrInitialStack;
		USHORT    OffsetPcrStackLimit;

		USHORT    OffsetPrcbPcrPage;
		USHORT    OffsetPrcbProcStateSpecialReg;
		USHORT    GdtR0Code;
		USHORT    GdtR0Data;

		USHORT    GdtR0Pcr;
		USHORT    GdtR3Code;
		USHORT    GdtR3Data;
		USHORT    GdtR3Teb;

		USHORT    GdtLdt;
		USHORT    GdtTss;
		USHORT    Gdt64R3CmCode;
		USHORT    Gdt64R3CmTeb;

		ULONG64   IopNumTriageDumpDataBlocks;
		ULONG64   IopTriageDumpDataBlocks;

		// Longhorn addition

		ULONG64   VfCrashDataBlock;
#endif

	} KDDEBUGGER_DATA64, *PKDDEBUGGER_DATA64;

	NtSystemDebugControlFunc NtSystemDebugControl
		= (NtSystemDebugControlFunc) GetProcAddress( GetModuleHandle( "ntdll.dll"), "NtSystemDebugControl" );

	bool Init()
	{
		HANDLE hToken;
		if( !OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken) )
		{
			return false;
		}
		BOOL ret = SetPrivilege(hToken, SE_DEBUG_NAME, TRUE);
		CloseHandle(hToken);

		return ret != FALSE;
	}


	bool GetVersionBlock( DBGKD_GET_VERSION64& versionBlock )
	{
		NTSTATUS ret = NtSystemDebugControl
			(
			SysDbgGetVersion,
			NULL,
			0,
			&versionBlock,
			sizeof(versionBlock),
			NULL
			);
		return ret ==0 ;
	}

	template <typename T>
	ULONG ReadMemory( T* dest, const DWORD address )
	{
		return ReadMemory( dest, address, 1 );
	}

	template <typename T>
	ULONG ReadMemory( T* dest, DWORD address, ULONG count )
	{
		MEMORY_QUERY query;
		ULONG length = -1;
		query.Address = address;
		query.Data = dest;
		query.Length = sizeof(T) * count;

		NtSystemDebugControl(
			SysDbgReadMemory,
			&query,
			sizeof(query),
			NULL,
			0,
			&length
			);

		return length;
	}
}

// dt -r nt!_MMPFN
#pragma pack(push,1)
struct _MMPFN
{
	DWORD dummy1;
	DWORD dummy2;
	DWORD dummy3;
	unsigned char dummy4;
	unsigned char type;
	unsigned short dummy5;
	DWORD dummy6;
	DWORD dummy7;
};

struct _MMPFN_PAE
{
	DWORD dummy1;
	DWORD dummy2;
	DWORD dummy3;
	unsigned char dummy4;
	unsigned char type;
	unsigned short dummy5;
	ULONGLONG dummy6;
	DWORD dummy7;
};
#pragma pack(pop)

template<typename T>
void CountupPages( __int64 counts[8], const SysDbg::KDDEBUGGER_DATA64& kddata )
{
	DWORD pfnDatabase = 0;
	SysDbg::ReadMemory( &pfnDatabase, static_cast<DWORD>( kddata.MmPfnDatabase ) );

	DWORD numberOfPhysicalPages = 0;
	SysDbg::ReadMemory( &numberOfPhysicalPages, static_cast<DWORD>( kddata.MmNumberOfPhysicalPages ) );

	T* frames = new T[numberOfPhysicalPages];

	SysDbg::ReadMemory( frames, pfnDatabase, numberOfPhysicalPages );

	for(DWORD i = 0; i<numberOfPhysicalPages; ++i)
	{
		const unsigned char index = frames[i].type & 0x07;
		++counts[index];
	}

	delete[] frames;
}

int main()
{
	if( !SysDbg::Init() )
	{
		return 1;
	}

	SysDbg::DBGKD_GET_VERSION64 KdVersionBlock;

	if( !SysDbg::GetVersionBlock( KdVersionBlock ) )
	{
		printf( "GetVersion64 falsed.\n" );
		return 1;
	}

	DWORD kddataAddr = 0;
	SysDbg::ReadMemory( &kddataAddr, static_cast<DWORD>( KdVersionBlock.DebuggerDataList ) );

	SysDbg::KDDEBUGGER_DATA64 kddata;
	SysDbg::ReadMemory( &kddata, kddataAddr );

	__int64 counts[8] = {0};
	const char *descs[8]= {
		"           Zeroed",
		"             Free",
		"          Standby",
		"         Modified",
		"Modified no-write",
		"              Bad",
		"           Active",
		"       Transition"
	};
	const int displayOrder[8] = {0,1,2,3,4,6,7,5};

	// NX bit requires PAE (Physical Address Extension)
	if( IsProcessorFeaturePresent( PF_PAE_ENABLED ) )
	{
		CountupPages<_MMPFN_PAE>( counts, kddata );
	}
	else
	{
		CountupPages<_MMPFN>( counts, kddata );
	}

	for( int i = 0; i < 8; ++i )
	{
		const int index = displayOrder[i];
		printf("%s : %8d MB\n", descs[index], 
			static_cast<int>( counts[index] * kddata.MmPageSize / ( 1024 * 1024 ) ) );
	}

	return 0;
}


// http://support.microsoft.com/kb/131065/en-us
BOOL SetPrivilege( 
				  HANDLE hToken,  // token handle 
				  LPCTSTR Privilege,  // Privilege to enable/disable 
				  BOOL bEnablePrivilege  // TRUE to enable. FALSE to disable 
				  )
{ 
	TOKEN_PRIVILEGES tp = { 0 }; 
	// Initialize everything to zero 
	LUID luid; 
	DWORD cb=sizeof(TOKEN_PRIVILEGES); 
	if(!LookupPrivilegeValue( NULL, Privilege, &luid ))
		return FALSE; 
	tp.PrivilegeCount = 1; 
	tp.Privileges[0].Luid = luid; 
	if(bEnablePrivilege) { 
		tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED; 
	} else { 
		tp.Privileges[0].Attributes = 0; 
	} 
	AdjustTokenPrivileges( hToken, FALSE, &tp, cb, NULL, NULL ); 
	if (GetLastError() != ERROR_SUCCESS) 
		return FALSE; 

	return TRUE;
}